from lange_.ap import AP
from lange_.gp import GP


class APwithBrackets:
    def __getitem__(self, item):
        """Helper class to allow overriding square brackets.

        Import object 'h' from module 'lange' to use it implicitly.

        Usage:
            >>> from lange_ import a_
            >>> (a_[0.6, 0.8, ..., 2])
            [0.6 0.8 .+. 2.0]
            >>> print(a_[0.6, 0.8, ..., 2])
            [0.6 0.8 1.0 1.2 1.4 1.6 1.8 2.0]
        """
        if isinstance(item, tuple):
            return AP(*item)
        else:
            return AP(item)


class GPwithBrackets:
    def __getitem__(self, item):
        """Helper class to allow overriding square brackets.

        Import object 'h_' from module 'lange' to use it implicitly.

        Usage:
            >>> from lange_ import g_
            >>> (g_[0.3, 0.6, ..., 2])
            [0.3 0.6 .*. 2.0]
            >>> print(g_[0.3, 0.6, ..., 2])
            [0.3 0.6 1.2]
        """
        if isinstance(item, tuple):
            return GP(*item)
        else:
            return GP(item)



import decimal as dec


def detect_precision(x, maxdigits=28):
    """
    Detect the precision (include digits before and after the decimal separator).

    The checking only works if 'max_digits' is large enough.

    Usage:
        >>> detect_precision(3)
        1
        >>> detect_precision(3.7)
        2
        >>> detect_precision(3.0)
        1
        >>> detect_precision(30.0e-4)
        4
        >>> detect_precision(30.0e4)
        6

    https://stackoverflow.com/questions/3018758/determine-precision-and-ret-of-particular-number-in-python
    """
    ctx = dec.Context()
    ctx.prec = maxdigits
    d1 = ctx.create_decimal(repr(float(x)))
    txt = format(d1, 'f')
    l = len(txt) - 1
    if int(x) == x:
        l -= 1
    return l



import decimal as dec
import math

from lange_.tricks import detect_precision


class Progression:
    def __init__(self, op, null, prod_f, div_f, pow_f, bins_f, args, maxdigits=28):
        self.prod_f, self.div_f, self.pow_f = prod_f, div_f, pow_f
        self.op = op
        self.decctx = dec.Context()
        null = self.decctx.create_decimal(null)
        if Ellipsis in args:
            if not (3 <= len(args) <= 4) or args[2] is not Ellipsis:
                raise Exception("Among 3 or 4 arguments, '...' should be the third.")

            # Detect level of precision and enforce it for the rest of this object life.
            end = 0 if len(args) == 3 else args[-1]
            self.precision = max(
                detect_precision(args[0], maxdigits),
                detect_precision(args[1], maxdigits),
                detect_precision(end, maxdigits)
            )
            self.decctx.prec = self.precision

            # Protect all provided numbers from floating point inequality issues (e.g. 0.8 - 0.6 != 0.2).
            self.start = self.decctx.create_decimal(args[0])
            self.second = self.decctx.create_decimal(args[1])
            self.end = self.decctx.create_decimal(math.inf if len(args) == 3 else args[-1])

            # Derived attributes.
            self.step = div_f(self.second, self.start)
            self.cast = int if all(isinstance(v, int) for v in args[:2]) and int(self.step) == self.step else float
            bins = bins_f(self.start, self.end, self.step).to_integral_exact(rounding=dec.ROUND_FLOOR)
            self.size = bins + self.decctx.create_decimal(1)

            def g():
                s, i = self.start, 0
                while i < self.size:
                    yield self.cast(s)
                    s = prod_f(s, self.step)
                    i += 1

            self.gen = g
        elif len(args) == 0:
            self.start, self.step, self.end = None, None, None
            self.size = 0
            self.gen = lambda: iter(())
            self.cast = lambda x: x
        elif 1 <= len(args) <= 2:
            self.start = self.decctx.create_decimal(args[0])
            self.step = null if len(args) == 1 else self.div_f(self.decctx.create_decimal(args[1]), self.start)
            self.end = self.decctx.create_decimal(args[-1])
            self.size = len(args)
            self.gen = lambda: iter(args)
            self.cast = int if all(isinstance(v, int) for v in args) and int(self.step) == self.step else float
        else:
            self.start, self.step, self.end = None, None, None
            self.size = len(args)
            self.gen = lambda: iter(args)
            self.cast = int if all(isinstance(v, int) for v in args) else float

    def __iter__(self):
        return self.gen()

    def __getitem__(self, item):
        if isinstance(item, slice):
            if self.start is None:
                return self.__class__(*list(self)[item])
            item_start = item.start or self.decctx.create_decimal(0)
            item_step = item.step or self.decctx.create_decimal(1)
            start = self.decctx.create_decimal(self[item_start])
            step = self.div_f(self.decctx.create_decimal(self[item_start + item_step]), start)
            second = self.prod_f(start, step)
            end = self.prod_f(start, self.pow_f(step, (item.stop - item_start - 1) // item_step))
            args = start, second, ..., end
            return self.__class__(*args)  # REMINDER: calling child class.

        if self.start is None:
            return self.__class__(list(self)[item])

        if not (0 <= item < self.size):
            raise Exception(f"Index {item} outside valid range [0; {self.size}[.")
        return self.cast(self.prod_f(self.start, self.pow_f(self.step, item)))

    def __str__(self):
        return f"[{' '.join(map(str, self))}]"

    def __repr__(self):
        if self.start is None:
            return str(self)
        end = self.cast(self.end)
        return f"[{self.cast(self.start)} {self.cast(self.prod_f(self.start, self.step))} .{self.op}. {end}]"

    def __invert__(self):
        return list(self)



# Geometric Progression

# "Forbidden" syntax.
import lange
print(~[0.4, 0.8, ..., 2])
# ...

# Conservative syntax.
from lange_ import g_
print(g_[0.4, 0.8, ..., 2])
# ...

pr = g_[0.4, 0.8, ...]
print(pr[:5])
# ...



# Arithmetic Progression

# "Forbidden" syntax.
import lange
print(-[0.6, 0.8, ..., 2])
# ...

# Conservative syntax.
from lange_ import a_
print(a_[0.6, 0.8, ..., 2])
# ...

pr = a_[0.6, 0.8, ...]
print(pr[:5])
# ...
